#!/usr/local/bin/python
#
# Name        :  CPUProfileDetailsParser.py
# Purpose     :  Script used to extract CPU usage summary data from the
#                built-in HPROF capabilities (used by Hadoop).  Due
#                to execution time and time-out issues, the sampling
#                layout is assumed.
#
#                Complexity is introduced in analyzing the profiles 
#                in terms of associating profiling output with the
#                application.  This information can be gleaned from
#                the combination of Job History -- which identifies
#                the job (job_<jobtracker-start>_<job#>) -- and lists
#                the worst performing tasks, which may or may not be
#                those sampled during profiling (depending on options).
#                The naming convention used for task profile output files
#                is:
#
#                  attempt_<jobtracker-start>_<job#>\
#                         _<type>_<task#>_<attempt#>.profile
#
#                where 
#                  jobtracker-start = timestamp of starting JobTracker
#                  job#             = the sequence number of the profiled job
#                  type             = "m" for map task, "r" for reduce task
#                  task#            = sequence number of the task
#                  attempt#         = sequence number of the task attempt
#                with zero-offset sequence numbers.
# Created     :  March 24, 2009
# Last Updated:  September 28, 2009
#
# Usage       :  cat <file-list> | python CPUProfileDetailsParser.py > \
#                   <outfilename>
#
# Copyright (c) 2009, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory.
# LLNL-CODE-416230
# All rights reserved.
#
# This file is part of the Data-Intensive Benchmark Suite.
# Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
# distribution.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License (as published by
# the Free Software Foundation) version 2.1 dated February 1999.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
# conditions of the GNU Lesser General Public License for more details.
#
# You should have recieved a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

import string

#
#####################################################

FALSE = 0
TRUE  = 1
SEP = ';'
# Note:  Count = Times Trace Active WHEN SAMPLING
#        Count = # Times method called WHEN "cpu=timing"
TITLES = [
  'File #',
  'Date',
  'Samples (#)',
  'Rank',
  'CPU Usage',
  'CPU Accum',
  'Times Trace Active (#)',
  'Trace Id',
  'Method',
]

def intToStr(intVal):
  return '%s' % intVal

def parseBegin(str):
  totalStr = ''
  openInd = str.find('(')
  if (openInd >= 18):
    endInd = str.find(')')
    totalStr = str[openInd+8:endInd]
    dateStr = str[endInd+1:len(str)-2]
  return [totalStr, dateStr]

def parseCPUDetails(line):
  vals = line.split()
  numVals = len(vals)
  if (numVals != 6):
    vals = null
  return vals


class CPUProfileDetails:

  def __init__(self):
    self.seq = -1
    self.total = ''
    self.date = ''
    self.rank = ''
    self.usage = ''
    self.accum = ''
    self.count = ''
    self.trace = ''
    self.method = ''
    print SEP.join(TITLES)

  def initData(self):
    self.total = ''
    self.date = ''
    self.rank = ''
    self.usage = ''
    self.accum = ''
    self.count = ''
    self.trace = ''
    self.method = ''

  def writeData(self):
    base   = intToStr(self.seq) + SEP + self.date + SEP + self.total + SEP
    detail = self.rank + SEP + self.usage + SEP + self.accum + SEP
    detail += self.count + SEP + self.trace + SEP
    print base + detail + self.method

  def parseStart(self, line):
    self.seq += 1
    self.initData()
    res = parseBegin(line)
    if ( (res != None) and (len(res) == 2) ):
      self.total = res[0]
      self.date  = res[1]

  def parseCPU(self, line):
    res = parseCPUDetails(line)
    if ( (res != None) and (len(res) == 6) ):
      self.rank = res[0]
      self.usage = res[1]
      self.accum = res[2]
      self.count = res[3]
      self.trace = res[4]
      self.method = res[5]


#
#####################################################
#                     MAIN
#####################################################
#
if __name__ == '__main__':
  import sys
  profile = CPUProfileDetails()
  extractData = FALSE
  try:
    for line in sys.stdin:
      if (line != None):
        if (line.startswith('CPU SAMPLES BEGIN')):
          profile.parseStart(line)
          extractData = TRUE
        elif (line.startswith('rank')):
          extractData = TRUE
        elif (line.startswith('CPU SAMPLES END')):
          extractData = FALSE
        elif (extractData):
          profile.parseCPU(line)
          profile.writeData()
  except:
    (type, value, trace) = sys.exc_info()
    import traceback
    traceback.print_exception(type, value, trace)
0
