#!/usr/local/bin/python
#
# Name        :  CPUProfileParser.py
# Purpose     :  Script used to extract CPU usage summary data from the
#                built-in HPROF capabilities (used by Hadoop).  Due
#                to execution time and time-out issues, the sampling
#                layout is assumed.
#
#                Complexity is introduced in analyzing the profiles 
#                in terms of associating profiling output with the
#                application.  This information can be gleaned from
#                the combination of Job History -- which identifies
#                the job (job_<jobtracker-start>_<job#>) -- and lists
#                the worst performing tasks, which may or may not be
#                those sampling during profiling (depending on options).
#                The naming convention used for task profile output files
#                is:
#
#                    attempt_<jobtracker-start>_<job#>\
#                           _<type>_<task#>_<attempt#>.profile
#
#                where 
#                  jobtracker-start = timestamp of starting JobTracker
#                  job#             = the sequence number of the profiled job
#                  type             = "m" for map task, "r" for reduce task
#                  task#            = sequence number of the task
#                  attempt#         = sequence number of the task attempt
#                with zero-offset sequence numbers.
# Created     :  March 24, 2009
# Last Updated:  September 28, 2009
#
# Usage       :  cat <file-list> | python CPUProfileParser.py > <outfilename>
#
# Copyright (c) 2009, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory.
# LLNL-CODE-416230
# All rights reserved.
#
# This file is part of the Data-Intensive Benchmark Suite.
# Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
# distribution.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License (as published by
# the Free Software Foundation) version 2.1 dated February 1999.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
# conditions of the GNU Lesser General Public License for more details.
#
# You should have recieved a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

import string
#

import string

#
#####################################################

FALSE = 0
TRUE  = 1
DETAILS = FALSE
SEP = ';'
# Note:  Count = Times Trace Active WHEN SAMPLING
#        Count = $ Times method called WHEN "cpu=timing"
TITLES = [
  'Sequence #',
  'Date',
  'Sample Size',
  'Method',
  'Mean Rank',
  'Mean CPU Usage (%)',
  'Mean Times Trace Active (#)',
]

def numToStr(val):
  return '%s' % val

def parseBegin(str):
  totalStr = ''
  openInd = str.find('(')
  if (openInd >= 18):
    endInd = str.find(')')
    totalStr = str[openInd+8:endInd]
    dateStr = str[endInd+1:len(str)-2]
  return [totalStr, dateStr]

def parseCPUDetails(line):
  vals = line.split()
  numVals = len(vals)
  if (numVals != 6):
    vals = null
  return vals


class CPUProfile:

  def __init__(self):
    self.seq = -1
    self.date = ''
    self.sampleSize = []
    self.details = {}
    self.summary = {}
    print SEP.join(TITLES)

  def initData(self):
    self.date = ''
    self.details.clear()

  def parseStart(self, line):
    self.seq += 1
    self.initData()
    res = parseBegin(line)
    if ( (res != None) and (len(res) == 2) ):
      self.date = res[1]
      self.sampleSize.append(int(res[0]))

  def parseCPU(self, line):
    res = parseCPUDetails(line)
    if ( (res != None) and (len(res) == 6) ):
      # Number of entries, Rank, Usage (%), Times Trace Active
      usage = res[1]
      entry = [ 1, float(res[0]), float(usage[:len(usage)-1]), float(res[3]) ]
      methName = res[5]
      #
      # First load the detail information
      if (self.details.has_key(methName)):
        vals = self.details.get(methName)
        for i in range(len(entry)):
          vals[i] += entry[i]
      else:
        self.details[methName] = entry
      #
      # Now load the summary information
      vals = self.summary.get(methName)
      if (self.summary.has_key(methName)):
        for i in range(len(entry)):
          vals[i] += entry[i]
      else:
        self.summary[methName] = entry
    else:
      print 'ERROR: parseCPU: Expected 6 items from parseCPUDetails()'

  def writeDetail(self):
    seq = self.seq
    base = numToStr(seq) + SEP + self.date + SEP 
    base += numToStr(self.sampleSize[seq]) + SEP
    for key in self.details.keys():
      vals = self.details.get(key)
      totals = key
      for ind in range(len(vals)):
        if (ind > 0):
          avg = vals[ind]/vals[0]
          totals += SEP + numToStr(avg)
      print base + totals 

  def writeSummary(self):
    print ''
    sum = 0
    for num in self.sampleSize:
      sum += num
    avgSize = float(sum)/float(self.seq)
    base   = SEP + 'Overall' + SEP + numToStr(avgSize) + SEP
    for key in self.summary.keys():
      vals = self.summary.get(key)
      totals = key 
      for ind in range(len(vals)):
        if (ind > 0):
          avg = float(vals[ind])/float(vals[0])
          totals += SEP + numToStr(avg)
      print base + totals 


#
#####################################################
#                     MAIN
#####################################################
#
if __name__ == '__main__':
  import sys
  profile = CPUProfile()
  extractData = FALSE
  try:
    for line in sys.stdin:
      if (line != None):
        if (line.startswith('CPU SAMPLES BEGIN')):
          profile.parseStart(line)
          extractData = TRUE
        elif (line.startswith('rank')):
          extractData = TRUE
        elif (line.startswith('CPU SAMPLES END')):
          extractData = FALSE
          if (DETAILS):
            profile.writeDetail()
        elif (extractData):
          profile.parseCPU(line)
    profile.writeSummary()
  except:
    (type, value, trace) = sys.exc_info()
    import traceback
    traceback.print_exception(type, value, trace)
0
