#!/usr/local/bin/python
#
# Name        :  JobHistoryParser.py
# Purpose     :  Script used to extract timing summary data from the
#                Job History report generated by executing
#                   $ ./bin/hadoop job -history <HDFS Log Directory>
# Created     :  February 13, 2009
# Last Updated:  September 28, 2009
#
# Usage       :  cat <file-list> | python JobHistoryParser.py > <outfilename>
#
# Copyright (c) 2009, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory.
# LLNL-CODE-416230
# All rights reserved.
#
# This file is part of the Data-Intensive Benchmark Suite.
# Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
# distribution.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License (as published by
# the Free Software Foundation) version 2.1 dated February 1999.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
# conditions of the GNU Lesser General Public License for more details.
#
# You should have recieved a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

import string
#

import string

#
#####################################################

FALSE = 0
TRUE  = 1
SEP = ';'
TITLES = [
  'Job Id',
  'Launch (sec)',
  'Task Setup (sec)',
  'Task Map (sec)',
  'Task Reduce (sec)',
  'Task Cleanup (sec)',
  'Elapsed (sec)',
  'Submitted',
  'Launched',
  'Setup',
  'Map',
  'Reduce',
  'Cleanup',
  'Finished',
  '# Mappers',
  '# Reducers',
]

def intToStr(intVal):
  return '%s' % intVal

def parseDate(str):
  dateStr = ''
  openInd = str.find('(')
  if (openInd > 18):
    dateStr = str[:openInd-2]
  return dateStr

def parseElapsed(str):
  sec = 0
  if (str != None):
    openInd = str.find('(')
    endInd = str.find(')')
    if ( (openInd > -1) and (endInd > openInd) ):
      timeStr = str[openInd+1:endInd]
      #print "DEBUG: parseElapsed: timeStr=", timeStr
      if (timeStr != None):
        minInd = timeStr.find('mins')
        commaInd = str.find(',')
        minEndInd = -1
        minVal = 0
        if (minInd > 0):
          minStr = timeStr[:minInd]
          minEndInd = minInd+6
          #print "DEBUG: parseElapsed: minStr=", minStr
          if ( (minStr != None) and (minStr != '') ):
            sec = 60 * int(minStr)
        secInd = timeStr.find('sec')
        if (secInd > -1):
          secStr = ''
          if (minEndInd > -1):
            secStr = timeStr[minEndInd:secInd]
          else:
            secStr = timeStr[:secInd]
          #print "DEBUG: parseElapsed: secStr=", secStr
          if ( (secStr != None) and (secStr != '') ):
            sec = sec + int(secStr)
  return sec

def parseTaskSummary(line):
  num = 0
  start = ''
  finish = ''
  sec = 0
  vals = line.split()
  numVals = len(vals)
  num = vals[2]  # Want number of successful tasks
  if (numVals > 5):
    # Want the start timestamp of the first task
    start = vals[5] + ' ' + vals[6]
    # Want the elapsed time over all such tasks
    elapsedStr = vals[9]
    if (numVals == 11):
      elapsedStr += ' ' + vals[10]
    finish = parseDate(elapsedStr)
    sec = parseElapsed(elapsedStr)
  return [ num, start, finish, sec ]


class JobHistory:

  def __init__(self):
    self.init = FALSE
    self.id = ''
    self.submit = ''
    self.launch = ''
    self.launchTime = 0
    self.finish = ''
    self.elapsedTime = 0
    self.setup = ''
    self.setupTime = 0
    self.map = ''
    self.mapTime = 0
    self.reduce = ''
    self.reduceTime = 0
    self.cleanup = ''
    self.cleanupTime = 0
    self.mappers = 0
    self.reducers = 0
    print SEP.join(TITLES)

  def writeData(self):
    if (self.init):
      jt = intToStr(self.launchTime) + SEP
      jt += intToStr(self.setupTime) + SEP
      jt += intToStr(self.mapTime) + SEP
      jt += intToStr(self.reduceTime) + SEP
      jt += intToStr(self.cleanupTime) + SEP
      jt += intToStr(self.elapsedTime) 
      ji = self.submit + SEP + self.launch + SEP + self.setup + SEP
      ji += self.map + SEP + self.reduce + SEP + self.cleanup + SEP 
      ji += self.finish
      jn = intToStr(self.mappers) + SEP + intToStr(self.reducers)
      print self.id + SEP + jt + SEP + ji + SEP + jn

  def parseId(self, line):
    self.init = TRUE
    self.id = line[12:len(line)-1]

  def parseSubmitted(self, line):
    self.submit = line[14:len(line)-1]

  def parseLaunched(self, line):
    str = line[13:len(line)-1]
    self.launch = parseDate(str)
    self.launchTime = parseElapsed(str)

  def parseFinished(self, line):
    str = line[13:len(line)-1]
    self.finish = parseDate(str)
    self.elapsedTime = parseElapsed(str)

  def parseSetup(self, line):
    res = parseTaskSummary(line)
    if ( (res != None) and (len(res) == 4) ):
      self.setup = res[1]
      self.setupTime = res[3]
    

  def parseMap(self, line):
    res = parseTaskSummary(line)
    if ( (res != None) and (len(res) == 4) ):
      self.mappers = res[0]
      self.map = res[1]
      self.mapTime = res[3]

  def parseReduce(self, line):
    res = parseTaskSummary(line)
    if ( (res != None) and (len(res) == 4) ):
      self.reducers = res[0]
      self.reduce = res[1]
      self.reduceTime = res[3]

  def parseCleanup(self, line):
    res = parseTaskSummary(line)
    if ( (res != None) and (len(res) == 4) ):
      self.cleanup = res[1]
      self.cleanupTime = res[3]


#
#####################################################
#                     MAIN
#####################################################
#
if __name__ == '__main__':
  import sys
  history = JobHistory()
  try:
    for line in sys.stdin:
      if (line != None):
        if (line.startswith('Hadoop job: ')):
          history.writeData()
          history.parseId(line)
        elif (line.startswith('Submitted At: ')):
          history.parseSubmitted(line)
        elif (line.startswith('Launched At: ')):
          history.parseLaunched(line)
        elif (line.startswith('Finished At: ')):
          history.parseFinished(line)
        elif (line.startswith('Setup')):
          history.parseSetup(line)
        elif (line.startswith('Map')):
          history.parseMap(line)
        elif (line.startswith('Reduce')):
          history.parseReduce(line)
        elif (line.startswith('Cleanup')):
          history.parseCleanup(line)
    history.writeData()
  except:
    (type, value, trace) = sys.exc_info()
    import traceback
    traceback.print_exception(type, value, trace)
0
