#!/usr/local/bin/python
#
# Name        :  JobNodeParser.py
# Purpose     :  Script used to extract node data from the raw Job History
#                data generated during Map-Reduce in Hadoop.  The file
#                is the one used by Hadoop's Job History report and is
#                buried in the HDFS log directory within the Map-Reduce 
#                job's output directory.
# Created     :  February 20, 2009
# Last Updated:  September 28, 2009
#
# Usage       :  cat <file-list> | python JobNodeParser.py > <outfilename>
#
# Copyright (c) 2009, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory.
# LLNL-CODE-416230
# All rights reserved.
#
# This file is part of the Data-Intensive Benchmark Suite.
# Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
# distribution.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License (as published by
# the Free Software Foundation) version 2.1 dated February 1999.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
# conditions of the GNU Lesser General Public License for more details.
#
# You should have recieved a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

import string
#

import string

#
#####################################################

FALSE = 0
TRUE  = 1
SEP = ';'
MHEADINGS = [
  'Job Id',
  'Map Task Id',
  'Mapper Host',
  'Mapper Node'
]
THEADINGS = [
  'Job Id',
  'Task Id',
  'Split Host',
  'Split Node'
]
NHEADINGS = [
  'Node',
  'Mapper?',
  'Use Count'
]

def intToStr(intVal):
  return '%s' % intVal

def extractTaskId(str):
  id = ''
  idInd = str.find('TASKID')
  if (idInd > 0):
    rem = str[idInd+8:]
    qInd = rem.find('"')
    if (qInd > 0):
      id = rem[:qInd]
  return id

def extractNode(str):
  parts = str.split('/')
  return parts[len(parts)-1]


class JobHistoryNode:

  def __init__(self):
    self.jobId = ''
    self.jobNodes = {}
    self.taskNodes = {}

  def writeData(self):
    if (self.jobId != ''):
      nodeList = {}
      mapperList = {}
      #
      # First write the data to output
      print SEP.join(MHEADINGS)
      for key in self.jobNodes.keys():
        host = self.jobNodes.get(key)
        node = extractNode(host)
        if (nodeList.has_key(node)):
          nodeList[node] = nodeList.get(node) + 1
        else:
          nodeList[node] = 1
          mapperList[node] = 'TRUE'
        print self.jobId + SEP + key + SEP + host + SEP + node

      print ''
      print SEP.join(THEADINGS)
      for key in self.taskNodes.keys():
        hostList = self.taskNodes.get(key)
        for host in hostList:
          node = extractNode(host)
          if (nodeList.has_key(node)):
            nodeList[node] = nodeList.get(node) + 1
          else:
            nodeList[node] = 1
            mapperList[node] = 'FALSE'
          print self.jobId + SEP + key + SEP + host + SEP + node

      print ''
      print SEP.join(NHEADINGS)
      for node in nodeList.keys():
        count = nodeList.get(node)
        if (count != None):
          isMapper = mapperList.get(node, 'FALSE')
          print node + SEP + isMapper + SEP + intToStr(count)
      print ''
      print "Total Nodes =", len(nodeList)
      print ''
      print ''
      #
      # Now clear the structures in case there are more job files 
      # being processed.
      self.jobId = ''
      self.jobNodes.clear()
      self.taskNodes.clear()

  def parseJobId(self, line):
    # Skip to the first character of the job name field
    rem = line[11:len(line)-1]
    if (rem != None):
      if (rem.find('JOBNAME') > 0):
        self.jobId = rem[:rem.find('"')]

  def parseMapAttempt(self, line):
    hostInd = line.find('HOSTNAME')
    if (hostInd > 0):
      taskId = extractTaskId(line[:hostInd])
      rem = line[hostInd+10:]
      qInd = rem.find('"')
      if (qInd > 0):
        self.jobNodes[taskId] = rem[:qInd]

  def parseTask(self, line):
    sInd = line.find('SPLITS')
    if (sInd > 0):
      taskId = extractTaskId(line[:sInd])
      rem = line[sInd+8:]
      qInd = rem.find('"')
      if (qInd > 0):
        list = rem[:qInd]
        if (list != None) and (list != ''):
          self.taskNodes[taskId] = list[:qInd].split(',')


#
#####################################################
#                     MAIN
#####################################################
#
if __name__ == '__main__':
  import sys
  history = JobHistoryNode()
  try:
    for line in sys.stdin:
      if (line != None):
        if (line.startswith('Meta')):
          history.writeData()
        if (line.startswith('Job JOBID="')):
          history.parseJobId(line)
        elif (line.startswith('MapAttempt')):
          history.parseMapAttempt(line)
        elif (line.startswith('Task TASKID')):
          history.parseTask(line)
    history.writeData()
  except:
    (type, value, trace) = sys.exc_info()
    import traceback
    traceback.print_exception(type, value, trace)
0
