#!/usr/local/bin/python
#
# Name        :  JobNodeParserSum.py
# Purpose     :  Script used to extract node data from the raw Job History
#                data generated during Map-Reduce in Hadoop.  The data
#                is maintained in the HDFS log directory associated with
#                the Map-Reduce output.
# Created     :  February 24, 2009
# Last Updated:  September 28, 2009
#
# Usage       :  cat <file-list> | python JobNodeParserSum.py > <outfilename>
#
# Copyright (c) 2009, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory.
# LLNL-CODE-416230
# All rights reserved.
#
# This file is part of the Data-Intensive Benchmark Suite.
# Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
# distribution.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License (as published by
# the Free Software Foundation) version 2.1 dated February 1999.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
# conditions of the GNU Lesser General Public License for more details.
#
# You should have recieved a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

import string
#

import string

#
#####################################################

FALSE = 0
TRUE  = 1
SEP = ';'
NAMEBASE = 'base-node-name'
NODEMAX = 40

def numToStr(val):
  return '%s' % val

def extractTaskId(str):
  id = ''
  idInd = str.find('TASKID')
  if (idInd > 0):
    rem = str[idInd+8:]
    qInd = rem.find('"')
    if (qInd > 0):
      id = rem[:qInd]
  return id

def extractNodeId(nodeBase, str):
  parts = str.split('/')
# Does the following extract the node number?!
  return parts[len(parts)-1][len(nodeBase):]


class JobHistoryNodeSum:

  def __init__(self, nodeBase, maxNodes):
    self.nodeBase = nodeBase
    self.maxNodes = maxNodes
    self.jobCount = 0
    self.nodeCountTotal = 0
    self.nodeTotal = []
    self.currJobId = ''
    self.jobNodes = {}
    self.taskNodes = {}

    headings = [ 'Job Id' ]
    for i in range(self.maxNodes):
      self.nodeTotal.append(0)
      headings.append('Node ' + numToStr(i+1))
    headings.append('Total Nodes')
    print SEP.join(headings)

  def writeData(self):
    if (self.currJobId != ''):
      self.jobCount += 1
      nodeList = []
      for i in range(self.maxNodes):
        nodeList.append(0)

      for key in self.jobNodes.keys():
        nodeId = extractNodeId(self.nodeBase, self.jobNodes.get(key))
        if nodeId.isdigit():
          nodeList[int(nodeId)-1] += 1
        else:
          print "Extracted non-numeric node id=", nodeId, "from jobNodes"

      for key in self.taskNodes.keys():
        hostList = self.taskNodes.get(key)
        for host in hostList:
          nodeId = extractNodeId(self.nodeBase, host)
          if nodeId.isdigit():
            nodeList[int(nodeId)-1] += 1
          else:
            print "Extracted non-numeric node id=", nodeId, "from taskNodes"

      outStr = self.currJobId + SEP
      nodeCount = 0
      for i in range(self.maxNodes):
        count = nodeList[i]
        outStr += numToStr(count) + SEP
        if (count > 0):
          nodeCount += 1
          self.nodeTotal[i] += count
      outStr += numToStr(nodeCount)
      self.nodeCountTotal += nodeCount

      print outStr

      # Clear the structures in case there are more job files 
      # being processed.
      self.currJobId = ''
      self.jobNodes.clear()
      self.taskNodes.clear()

  def writeAverage(self):
      outStr = 'Node Average' + SEP
      for i in range(self.maxNodes):
        avg = self.nodeTotal[i] / self.jobCount
        outStr += numToStr(avg) + SEP
      avg = self.nodeCountTotal / self.jobCount
      print outStr + numToStr(avg)

  def parseJobId(self, line):
    # Skip to the first character of the job name field
    rem = line[11:len(line)-1]
    if (rem != None):
      if (rem.find('JOBNAME') > 0):
        self.currJobId = rem[:rem.find('"')]

  def parseMapAttempt(self, line):
    hostInd = line.find('HOSTNAME')
    if (hostInd > 0):
      taskId = extractTaskId(line[:hostInd])
      rem = line[hostInd+10:]
      qInd = rem.find('"')
      if (qInd > 0):
        self.jobNodes[taskId] = rem[:qInd]

  def parseTask(self, line):
    sInd = line.find('SPLITS')
    if (sInd > 0):
      taskId = extractTaskId(line[:sInd])
      rem = line[sInd+8:]
      qInd = rem.find('"')
      if (qInd > 0):
        list = rem[:qInd]
        if (list != None) and (list != ''):
          self.taskNodes[taskId] = list[:qInd].split(',')


#
#####################################################
#                     MAIN
#####################################################
#
if __name__ == '__main__':
  import sys
  history = JobHistoryNodeSum(NAMEBASE, NODEMAX)
  try:
    for line in sys.stdin:
      if (line != None):
        if (line.startswith('Meta')):
          history.writeData()
        if (line.startswith('Job JOBID="')):
          history.parseJobId(line)
        elif (line.startswith('MapAttempt')):
          history.parseMapAttempt(line)
        elif (line.startswith('Task TASKID')):
          history.parseTask(line)
    history.writeData()
    history.writeAverage()
  except:
    (type, value, trace) = sys.exc_info()
    import traceback
    traceback.print_exception(type, value, trace)
0
