/**
 * Name   :  AttackClassification.java
 * Purpose:  Attack classification information for a single HTTP request.
 *           It is expected that the request identifier and/or request text
 *           will be supplied, though not both are required.
 * Author :  T. Dahlgren
 * Created:  January 7, 2009
 *
 * Acknowledgments:  
 *   This class mimics Brian Gallagher's non-Hadoop implementation's output.
 *     
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package gov.llnl.hadoop.classifier.httpRequest;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.lang.Double;
import java.lang.Object;
import java.lang.String;
import java.lang.StringBuffer;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Map;
import org.apache.hadoop.io.WritableComparable;

public class AttackClassification implements WritableComparable {

    /* ====================== ATTRIBUTES ====================== */
    private boolean g_typeOnly = true;

    /*
     * d_initialized TRUE if initialization phase completed. d_requestId
     * Identifier associated with the request. d_predClass Predicted
     * classification. d_dist Calculated distribution reflecting probability of
     * each attack type. d_request Full text of the request. d_format Format
     * used when outputting decimal numbers.
     * 
     * For simplicity, mirroring the attribute ordering in Brian's code to
     * reflect the ordering in his model. (The expectation/hope is to re-use his
     * model for initial testing.)
     */
    private boolean d_initialized = false;
    private String d_requestId = null;
    private String d_predClass = null;
    private Map<String, Double> d_dist = null;
    private String d_request = null;

    private static final DecimalFormat d_format = new DecimalFormat("0.000000");

    public AttackClassification() {
    } // AttackClassification()

    public AttackClassification(String id, String type,
            Map<String, Double> dist, String req) {
        d_requestId = id;
        d_predClass = type;
        d_dist = new HashMap<String, Double>(dist);
        d_request = req;
        d_initialized = true;
    } // AttackClassification(String, String, Map, String)

    @SuppressWarnings("unchecked")
    public void readFields(DataInput in) throws IOException {
        ObjectInputStream inStream = new ObjectInputStream((InputStream) in);

        try {
            d_requestId = (String) inStream.readObject();
            d_predClass = (String) inStream.readObject();
            d_dist = (Map<String, Double>) inStream.readObject();
            d_request = (String) inStream.readObject();
            d_initialized = true;
        } catch (ClassNotFoundException cnfe) {
            System.err.println("Unable to populate attack classification "
                    + "information from the input stream.");
            cnfe.printStackTrace();
        } finally {
            inStream.close();
        }
    } // readFields(DataInput): void

    public void write(DataOutput out) throws IOException {
        ObjectOutputStream outStream = new ObjectOutputStream(
                (OutputStream) out);

        try {
            outStream.writeObject(d_requestId);
            outStream.writeObject(d_predClass);
            outStream.writeObject(d_dist);
            outStream.writeObject(d_request);
        } finally {
            outStream.close();
        }
    } // write(DataOutput)

    /*
     * Generate stringified version of classification information to match Brian
     * Gallagher's output.
     */
    public String toString() {
        StringBuffer str = new StringBuffer();
        String reqIdStr = (d_requestId != null) ? d_requestId + " " : "";
        if (d_initialized) {
            if (!g_typeOnly) {
                str.append("\n--------------------------\n");
                str.append(reqIdStr + d_request + "\n");
                for (Map.Entry<String, Double> e : d_dist.entrySet()) {
                    str.append("P(" + e.getKey() + ") = "
                            + d_format.format(e.getValue()) + "\n");
                }
                str.append("Predicted type: ");
            }
            str.append(d_predClass);
        }
        return str.toString();
    } // toString(): String

    /*
     * Comparison function to allow object sorting (as in secondary sort!) Don't
     * really have anything worth comparing here, so simply returning 0 to
     * indicate relative order in "sorted" list does not matter.
     */
    public int compareTo(Object obj) {
        return 0;
    } // compareTo(Object): int
} // AttackClassification
