/**
 * Name   :  TFIDFAttackClassifierV1.java
 * Purpose:  This is a Hadoop Map/Reduce implementation of a class
 *           used to classify HTTP requests using a TF-IDF weighting
 *           model.
 *           An instance of the class processes a text file containing
 *           one request per line. Output consists of a classification 
 *           for each request based on an attack model read from a 
 *           binary file.
 * Author :  T. Dahlgren
 * Created:  January 7, 2009
 *
 * Usage  :  ${HADOOP_BIN}/hadoop jar ${HTFIDF_JAR} TFIDFAttackClassifierV1 \
 *              numMaps inDir modelFile outDir hasRequestIds
 *     
 *           where 
 *              numMaps is the number of mapper tasks
 *                  0 => Do not set the number
 *                 -1 => Set based on cluster size
 *                 >0 => Request the specified number of mappers
 *              inDir is location in HDFS of the requests file(s)
 *              modelFile is location in HDFS of the TF-IDF Term Data file
 *              outDir is location in HDFS for output (including logs)
 *              hasRequestIds is 'true' if each line of the[/each] input
 *              file starts with the request identifier
 *    
 * WARNING:  EACH run requires a new output directory.  Otherwise, you will
 *    get a FileAlreadyExistsException.
 *
 * Acknowledgments:  
 *   This class is a Hadoop Map/Reduce implementation of Brian Gallagher's
 *   TFIDFClassifier for HTTP attacks.
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package gov.llnl.hadoop.classifier.httpRequest;

import gov.llnl.hadoop.classifier.httpRequest.TFIDFAttackModelV1;
import gov.llnl.hadoop.classifier.httpRequest.AttackClassification;

import java.io.IOException;
import java.lang.Exception;
import java.lang.Integer;
import java.lang.String;
import java.util.Iterator;
import java.util.Map;

import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.ClusterStatus;
import org.apache.hadoop.mapred.Counters;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.RunningJob;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

public class TFIDFAttackClassifierV1 extends Configured implements Tool {
    /* ====================== ATTRIBUTES ====================== */
    /*
     * The "better" way to handle profiling is to set the properties
     * in ${HADOOP}/conf/hadoop-site.xml.  Specifically, set
     * mapred.task.profile to value "true" and the desired map and
     * reduce task ranges in the mapred.task.profile.maps and 
     * mapred.task.profile.reduces properties.
     */
/*
    public static final boolean g_doProfiling = false;
*/
    public static float d_configSec = 0.0f;

    /*
     * The following attributes are literals for the names of properties to be
     * shared across map/reduce tasks.
     * 
     * g_lt_myName Name of this classifier (for display purposes).
     * g_prop_modelFile Name of the model file within the HDFS.
     */
    public static final String g_lt_myName = "TFIDFAttackClassifierV1";
    public static final String g_prop_modelFile 
                               = "TFIDFAttackClassifierV1.modelFile";
    public static final String g_prop_extractRequestId
                               = "TFIDFAttackClassifierV1.extractRequestIds";
/*
    public static final String g_prop_profileOptions
                               = "mapred.task.profile.params";
    public static final String g_prop_profileMaps
                               = "mapred.task.profile.maps";
    public static final String g_prop_profileReduces
                               = "mapred.task.profile.reduces";
*/

    public static final int g_numMapsPerHost = 3; // # cores (=2) + 1 

/*
 * Initial timing...
    private enum Counter { GetClassMSec, GetClassIters, MapMSec, ReduceMSec };
*/
 

    /* ====================== INNER CLASSES ====================== */
    /*
     * Map requests to Text.
     */
    public static class AttackMapper extends MapReduceBase implements
            Mapper<LongWritable, Text, Text, AttackClassification>
    {
        /* ====================== ATTRIBUTES ====================== */
        /*
         * d_model     = In-memory representation of the attack model. 
         * d_extractId = TRUE if the request identifier is the first 
         *               string in the value (with tabs assumed between 
         *               the request identifier and the request text.
         */
        private TFIDFAttackModelV1 d_model = null;
        private boolean d_extractId = false;
/*
 * Initial timing...
        private long    d_startMSec = 0;
        private long    d_stopMSec = 0;
        private boolean d_incrModel = false;
*/


        /* ====================== METHODS ====================== */
        /*
         * Retrieve value of relevant global properties (set in run()).
         */
        public void configure(JobConf conf) {
/*
 * Initial timing...
            d_startMSec = System.currentTimeMillis();
*/
            String filename = conf.get(g_prop_modelFile);
            try {
                FileSystem fs = FileSystem.get(conf);
                Path file = new Path(filename);
                if (fs.isFile(file)) {
                    d_model = new TFIDFAttackModelV1(fs.open(file));
                } else {
                    System.err.println(g_lt_myName 
                            + ": AttackMapper: configure: "
                            + "Unable to load attack model in file named \""
                            + filename + "\" since it does not exist or is not "
                            + "a file in the file system.");
                }
            } catch (IOException ioe) {
                System.err.println(g_lt_myName + ": AttackMapper: configure: "
                        + "Unable to load the attack model in file named \""
                        + filename + "\" due to exception.");
                ioe.printStackTrace();
            }
/*
 * Initial timing...
            d_stopMSec = System.currentTimeMillis();
*/

            String propStr = conf.get(g_prop_extractRequestId);
            d_extractId = (propStr != null) ? propStr.equalsIgnoreCase("true")
                    : false;
        } // configure(JobConf)

        /*
         * Classify the request.  Each input split is assumed to be a 
         * single HTTP request.
         */
        public void map(LongWritable key, Text value,
                OutputCollector<Text, AttackClassification> output,
                Reporter reporter) throws IOException 
        {
/*
 * Initial timing...
            if (!d_incrModel) {
              reporter.incrCounter(Counter.GetClassMSec,
                                   d_stopMSec - d_startMSec);
              reporter.incrCounter(Counter.GetClassIters, 1);
              d_incrModel = true;
            }

            long startMapMSec = System.currentTimeMillis(); 
*/
 
            String valueStr = value.toString();
            String requestId = null;
            String request = null;
            AttackClassification myClass = null;

            if (d_extractId) {
                int tabInd = valueStr.indexOf("\t");
                if (tabInd > 0) {
                    requestId = valueStr.substring(0, tabInd);
                    request = valueStr.substring(tabInd + 1);
                } else {
                    System.err.println(g_lt_myName
                            + ": ERROR:  The following request line does NOT "
                            + "conform to <requestId>\t<requestText>:");
                    System.err.println("  " + valueStr);
                }
            } else {
                requestId = key.toString();
                request = valueStr;
            }

            /*
             * Determine which attack type has the highest probability for this
             * request.
             */
            if (request != null) {
                Map<String, Double> dist = d_model.calcDistribution(request);
                String attackType = null;
                double maxProb = 0.0;

                for (Map.Entry<String, Double> attack : dist.entrySet()) {
                    double aProb = attack.getValue().doubleValue();
                    if (aProb > maxProb) {
                        attackType = attack.getKey();
                        maxProb = aProb;
                    }
                }
                /*
                 * Is this sufficient reporting when profile sampling?
                 */
/*
                reporter.progress();
*/
                myClass = new AttackClassification(requestId, attackType, dist,
                        request);
                output.collect(new Text(requestId), myClass);
            } // endif have a request to process

/*
 * Initial timing...
            reporter.incrCounter(Counter.MapMSec,
                                 System.currentTimeMillis() - startMapMSec);
*/
        } // map(...)
    } // AttackMapper

    /**
     * Essentially a no op, emitting entries from
     * the mapper for output.
     */
    public static class AttackReducer extends MapReduceBase implements
            Reducer<Text, AttackClassification, Text, AttackClassification>
    {
        /* ====================== METHODS ====================== */
        public void reduce(Text key, Iterator<AttackClassification> values,
                OutputCollector<Text, AttackClassification> output,
                Reporter reporter) throws IOException 
        {
/*
 * Initial timing...
            long startReduceMSec = System.currentTimeMillis(); 
*/
 
            /*
             * Only want to output the classification information stored in each
             * value (though there should only be one based on existing source).
             */
            while (values.hasNext()) {
                output.collect(key, values.next());
            }

/*
 * Initial timing...
            reporter.incrCounter(Counter.ReduceMSec,
                                 System.currentTimeMillis() - startReduceMSec);
*/
        } // reduce(...)
    } // AttackReducer

    /* ====================== METHODS ====================== */
    static int printUsage() {
        System.out.println(g_lt_myName
                + " <#mappers> <inDir> <modelFile> <outDir> <haveRequestIds>");
        System.out.println("                  where");
        System.out.println("                    "
                + "<#mappers> is the number of map tasks");
        System.out.println("                    "
                + "  (0=do not set, <0=calculate it, >0=requested number);");
        System.out.println("                    "
                + "<inDir>, <modelFile>, and <outDir> are in the HDFS;");
        System.out.println("                    "
                + "<haveRequestIds> should be 'true' if each");
        System.out.println("                    "
                + "line of the input file(s) start with the associated");
        System.out.println("                    "
                + "identifier followed by a tab.");
        ToolRunner.printGenericCommandUsage(System.out);
        return -1;
    } // printUsage(): int

    /*
     * Classifies each request in the input path based on the specified model.
     */
    private void classifyRequests(int numMaps, String inPath, String modelFile,
            String outPath, String haveRequestIds) throws Exception {
/*
 * Initial timing...
        long startMSec = System.currentTimeMillis();
*/

        JobConf conf = new JobConf(getConf(), TFIDFAttackClassifierV1.class);
        conf.setJobName("TFIDFAttackClassifierV1-ClassifyRequests");

        int numMappers = 0;
        int numReducers = 0; // As long as reduce() is a no-op.
        if (numMaps > 0) {
          numMappers = numMaps;
        } else if (numMaps < 0) {
            /*
             * Calculate default number of maps and reduces.
             * (Based on Hadoop-0.19.0's Join.java example.)
             */
            JobClient client = new JobClient(conf);
            ClusterStatus cluster = client.getClusterStatus();
            numMappers= cluster.getTaskTrackers() * g_numMapsPerHost;
            //numReducers = (int) (cluster.getMaxReduceTasks() * 0.9);
        }
        conf.setNumMapTasks(numMappers);
        conf.setNumReduceTasks(numReducers);

        /*
         * Ensure input splits correspond to a single line of the input file.
         */
        FileInputFormat.setInputPaths(conf, new Path(inPath));
        conf.setInputFormat(TextInputFormat.class);

        conf.setOutputKeyClass(Text.class);
        conf.setOutputValueClass(AttackClassification.class);

        conf.setMapperClass(AttackMapper.class);
        conf.setReducerClass(AttackReducer.class);

        FileOutputFormat.setOutputPath(conf, new Path(outPath));
        conf.setOutputFormat(TextOutputFormat.class);

        /*
         * Establish custom property(ies) shared by map/reduce tasks,
         * such as the model file, request id extraction
         */
        conf.set(g_prop_modelFile, modelFile);
        conf.set(g_prop_extractRequestId, haveRequestIds);

        /*
         * Establish profiling.  Should default to include map and 
         * reduce task profiling.  Alternatively, these can/should 
         * be set in the site-specific configuration file.
         */
/*
        if (g_doProfiling) {
            conf.setProfileEnabled(true);
            conf.set(g_prop_profileOptions, 
                "-agentlib:hprof=cpu=samples,force=n,thread=n,"
              + "verbose=n,file=%s");
            conf.set(g_prop_profileMaps, "0-4");
            conf.set(g_prop_profileReduces, "0-4");
        }
*/

        /*
         * Finalize timing of configuration set up.
         */
/*
 * Initial timing...
        long elapsedMSec = System.currentTimeMillis() - startMSec;
        d_configSec = ((float)elapsedMSec) / ((float)1000.0);
*/

        /*
         * Finally, run the calculation phase.
         */
        JobClient.runJob(conf);

       /*
        * The following are unnecessary as the system automatically 
        * outputs them, but the approach is useful if were to 
        * perform iterations:
        *
        RunningJob rj = JobClient.runJob(conf);
        Counters counts = rj.getCounters();
        System.out.println("Classification (msec)    = " + elapsedMSec);
        System.out.println("...Mapper Time (msec)    = " 
                           + counts.getCounter(Counter.MapMSec));
        System.out.println("...Reducer Time (msec)   = " 
                           + counts.getCounter(Counter.ReduceMSec));
        */
    } // classifyRequests(String, String, String): void

    /**
     * Execute the TFIDFAttackClassifierV1 behavior by extracting requests from 
     * a text file; calculating the TF-IDF based on the default model (read 
     * from the data directory); and outputting the resulting classifications 
     * -- one per request.
     * 
     * Arguments: 
     *  args[0] = Number of map tasks
     *  args[1] = HDFS directory of input file(s) containing requests
     *              NOTE: Input files are expected to contain at least a 
     *              single HTTP request per line. Each line optionally starts 
     *              with the request identifier (see args[2]). 
     *  args[2] = HDFS file (directory + filename) for model file.
     *  args[3] = HDFS output directory for classification results. 
     *  args[4] = Boolean indicating whether the first (tab-delimited) field 
     *            of each line of the input file(s) is a request identifier.
     */
    public int run(String[] args) throws Exception {
        int res = 0;

        if (args.length == 5) {
            int numMaps = 0;
            try {
              numMaps = Integer.parseInt(args[0]);
            } catch (NumberFormatException exc) {
                System.out.println("ERROR:  Integer expected instead of " 
                                  + args[0]);
                return printUsage();
            }
            classifyRequests(numMaps, args[1], args[2], args[3], args[4]);
        } else {
            System.out.println(g_lt_myName
              + " <#mappers> <inDir> <modelFile> <outDir> <haveRequestIds>");
            System.out.println("                  where");
            System.out.println("                  "
              + "  <#mappers> is the number of map tasks (0=do not set, "
              + "<0=calculate it);");
            System.out.println("                  "
              + "  <inDir>, <modelFile>, and <outDir> are in the HDFS;");
            System.out.println("                  "
              + "  <haveRequestIds> should be 'true' if each");
            System.out.println("                  "
              + "  line of the input file(s) start with the associated");
            System.out.println("                  "
              + "  identifier followed by a tab.");
            System.out.println();
            ToolRunner.printGenericCommandUsage(System.out);
            res = -1;
        }

        return res;
    } // run(String[]): int

    /* ***************************************************************** */
    /* main() */
    /* ***************************************************************** */
    // @SuppressWarnings("unchecked")
    public static void main(String[] args) throws Exception {
/*
 * Initial timing...
        long startMSec = System.currentTimeMillis();
*/

        int res = ToolRunner.run(new Configuration(), 
                                 new TFIDFAttackClassifierV1(), args);

/*
 * Initial timing...
        long elapsedMSec = System.currentTimeMillis() - startMSec;
        float elapsedSec = ((float)elapsedMSec) / ((float)1000.0);
        System.out.println("Configuration (sec)      = " + d_configSec);
        System.out.println("Elapsed time (sec)       = " + elapsedSec);
*/

        System.exit(res);
    } // main(String[])

} // TFIDFAttackClassifierV1
