/**
 * Name   :  TFIDFAttackModelBuilderV1.java
 * Purpose:  This is a Hadoop Map/Reduce implementation of a class
 *           used to build a TF-IDF weighting model for classifying
 *           HTTP attacks.
 *           An instance of the class processes a text file containing
 *           one attack type, HTTP request per line and assumes tab 
 *           separation.  Output consists of binary representation of
 *           the attack model.
 * Author :  T. Dahlgren
 * Created:  January 9, 2009
 * Usage  :  ${HADOOP_BIN}/hadoop jar ${HTFIDF_JAR} TFIDFAttackModelBuilderV1 \
 *              inputDir modelFile stripIds
 *     
 *           where inputDir    is location in HDFS of the requests file. 
 *                 modelFile	  is the output file directory+name for the 
 *                             generated model file.
 *                 stripIds    is 'true' if request identifiers appear
 *                             in the first "column" of the input file(s),
 *                             so must be stripped.
 *
 * Acknowledgments:  
 *   This class is a Hadoop Map/Reduce implementation used to build a
 *   model in a manner similar to Brian Gallagher's TFIDFClassifier.
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package gov.llnl.hadoop.classifier.httpRequest;

import gov.llnl.hadoop.classifier.httpRequest.TFIDFAttackModelV1;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.Exception;
import java.lang.String;
import java.lang.StringBuffer;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FSDataOutputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

public class TFIDFAttackModelBuilderV1 extends Configured implements Tool {
    /* ====================== ATTRIBUTES ====================== */
    /*
     * The following attributes are literals for the names of properties to be
     * shared across map/reduce tasks.
     * 
     * g_lt_myName Name of this class (for display purposes). g_prop_modelFile
     * Name of the model file within the HDFS.
     */
    public static final String g_lt_myName = "TFIDFAttackModelBuilderV1";
    public static final String g_lt_midPath = "clean";
    public static final String g_prop_stripRequestIds 
			= "TFIDFAttackModelBuilderV1.stripRequestIds";

    /* ====================== INNER CLASSES ====================== */
    /*
     * Extract "[requestId\t]attackType\thttpRequest" entries from 
     * the input file(s) and map to <attackType, httpRequest> entries.
     */
    public static class InputMapper extends MapReduceBase implements
            Mapper<LongWritable, Text, Text, Text>
    {
        /* ====================== ATTRIBUTES ====================== */
        /*
         * d_stripIds TRUE if the request identifier is the first string in the
         * value so, therefore, must be removed from the request.
         */
        private boolean d_stripIds = false;

        /* ====================== METHODS ====================== */
        /*
         * Retrieve value of relevant global properties (set in
         * extractTrainingData()).
         */
        public void configure(JobConf conf) {
            d_stripIds = conf.get(g_prop_stripRequestIds).equalsIgnoreCase(
                    "true");
        } // configure(JobConf)

        /*
         * Map request entries from training set to an individual 
         * attack type and request.
         */
        public void map(LongWritable key, Text value,
                OutputCollector<Text, Text> output, Reporter reporter)
                throws IOException {
            String valueStr = value.toString();
            String startStr = null;
            String attackType = null;
            String request = null;
            String msg = "";
            int tabInd = -1;

            /*
             * Remove the request identifier, if (supposed to be) present.
             */
            if (d_stripIds) {
                msg = "<requestId>\t";
                tabInd = valueStr.indexOf("\t");
                if (tabInd > 0) {
                    startStr = valueStr.substring(tabInd + 1);
                }
            } else {
                startStr = valueStr;
            }

            /*
             * Extract the attack type and request.
             */
            if ((startStr != null) && (startStr.length() > 0)) {
                tabInd = startStr.indexOf("\t");
                if (tabInd > 0) {
                    attackType = startStr.substring(0, tabInd);
                    request = startStr.substring(tabInd + 1);
                }
            }

            /*
             * Emit the request input for processing with map/reduce.
             */
            if ((attackType != null) && (request != null)) {
                output.collect(new Text(attackType), new Text(request));
            } else {
                System.err.println(g_lt_myName
                        + ": ERROR:  The following request line does NOT "
                        + "conform to " + msg + "<attackType>\t<request"
                        + "Text> as expected:");
                System.err.println("  " + valueStr);
            } // endif have a (valid) request to process
        } // map(...)
    } // InputMapper

    /**
     * Combine all requests for a given attack type into a "single" request.
     */
    public static class InputCombiner extends MapReduceBase implements
            Reducer<Text, Text, Text, Text>
    {
        /* ====================== METHODS ====================== */
        public void reduce(Text key, Iterator<Text> values,
                OutputCollector<Text, Text> output, Reporter reporter)
                throws IOException {
            /*
             * Accumulate the requests for the given attack type (as Brian had
             * done off-line).
             */
            StringBuffer str = new StringBuffer();
            while (values.hasNext()) {
                str.append(values.next().toString());
            }

            output.collect(key, new Text(str.toString()));
        } // reduce(...)
    } // InputCombiner

    /**
     * Combine all requests for a given attack type into a "single" request, 
     * prepending the attack type (key).  Also build the model for the
     * given attack type.
     */
    public static class InputReducer extends MapReduceBase implements
            Reducer<Text, Text, Text, Text>
    {
        /* ====================== METHODS ====================== */
        public void reduce(Text key, Iterator<Text> values,
                OutputCollector<Text, Text> output, Reporter reporter)
                throws IOException {
            /*
             * Accumulate the requests for the given attack type (as Brian had
             * done off-line). Prepend the attack type key to the (final)
             * compound request so the general TextOutputFormat will output
             * attack type as well.
             */
            StringBuffer str = new StringBuffer();
            //str.append(key.toString() + "\t");
            //String keyStr = key.toString();
            while (values.hasNext()) {
                String valueStr = values.next().toString();
                str.append(valueStr);
            }

            /*
             * And, just for "completeness", output the aggregated requests
             * to the output file (for later comparison).
             */
            output.collect(key, new Text(str.toString()));
        } // reduce(...)
    } // InputReducer

    /* ====================== METHODS ====================== */
    /**
     * Read request sets from the input stream.
     */
    private Map<String, String> readRequestSets(InputStream in) {
        Map<String, String> requests = new HashMap<String, String>();

        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        try {
            String line = null;
            String attackType = null;
            String requestSet = null;
            while ((line = br.readLine()) != null) {
                int tabInd = line.indexOf("\t");
                if (tabInd > 0) {
                    attackType = line.substring(0, tabInd);
                    requestSet = line.substring(tabInd + 1);
                    requests.put(attackType, requestSet);
                }
            }
        } catch (IOException ioe) {
            System.err.println(g_lt_myName + "Unable to read request sets "
                    + "from the input stream.");
            ioe.printStackTrace();
        }

        return requests;
    } // readRequestSets(InputStream): Map<String, String>

    private void buildModel(FileSystem fs, Path inFilePath, 
                            TFIDFAttackModelV1 model) 
    {
        try {
            FSDataInputStream inFile = fs.open(inFilePath);
            Map<String, String> requestsIn = readRequestSets(inFile);
            model.buildModel(requestsIn);
        } catch (IOException exc) {
            System.err.println(g_lt_myName + ": buildModel: " 
                    + "Unable to build the attack model.");
            exc.printStackTrace();
        }
    } // buildModel(FileSystem, Path, TFIDFAttackModelV1): void

    private void writeModel(TFIDFAttackModelV1 model, FileSystem fs, 
                            Path modelFilePath) {
        try {
            FSDataOutputStream modelOut = fs.create(modelFilePath);
            model.write(modelOut);
        } catch (IOException exc) {
            System.err.println(g_lt_myName + ": writeModel: " 
                              + "Unable to output the attack model.");
            exc.printStackTrace();
        }
    } // writeModel(TFIDFAttackModelV1, FileSystem, Path): void

    /*
     * Use the input file created by Map/Reduce to create the attack model.
     */
    private void buildInMemoryModel(String inDir, String modelFile,
            String stripIds) throws Exception {
        JobConf conf = new JobConf(getConf(), TFIDFAttackModelBuilderV1.class);
        conf.setJobName("TFIDFAttackModelBuilderV1-buildInMemoryModel");
        conf.set(g_prop_stripRequestIds, stripIds);

        FileSystem fs = FileSystem.get(conf);

        Path inPath = new Path(inDir);
        Path intermedPath = new Path(inPath, new Path(g_lt_midPath));
        boolean inIsDir = fs.getFileStatus(inPath).isDir();

        Path modelFilePath = new Path(modelFile);
        boolean modelExists = fs.exists(modelFilePath);
        if (inIsDir && (!modelExists)) {
            /*
             * Ensure input splits correspond to a single line of the input
             * file.
             */
            FileInputFormat.setInputPaths(conf, inPath);
            conf.setInputFormat(TextInputFormat.class);

            /*
             * Establish <K, V> and other classes for Map/Reduce.
             */
            conf.setOutputKeyClass(Text.class);
            conf.setOutputValueClass(Text.class);

            conf.setMapperClass(InputMapper.class);
            conf.setCombinerClass(InputCombiner.class);
            conf.setReducerClass(InputReducer.class);

            /*
             * Leverage TextOutputFormat and the value's toString()
             * functionality to generate the desired Map/Reduce output.
             */
            FileOutputFormat.setOutputPath(conf, intermedPath);
            conf.setOutputFormat(TextOutputFormat.class);

            /*
             * Execute the map/reduce job to aggregate requests by attack type.
             */
            JobClient.runJob(conf);

            /*
             * Now use the equivalent of Brian's method to generate his TF-IDF
             * model but using the distributed file created above.
             */
            TFIDFAttackModelV1 model = new TFIDFAttackModelV1();
            Path inFilePath = new Path(intermedPath, new Path("part-00000"));
            buildModel(fs, inFilePath, model);
            writeModel(model, fs, modelFilePath);
        } else if (modelExists) {
            System.err.println(g_lt_myName + ": buildInMemoryModel: Unable to "
                    + "proceed since the model file called \"" + modelFile 
                    + "\" already exists.");
        } else /* if (!inDir) */ {
            System.err.println(g_lt_myName + ": buildInMemoryModel: Unable to "
                    + "proceed since the input directory does not exist in "
                    + "the file system.");
        }
    } // buildInMemoryModel(String, String, String): void

    /**
     * Execute the TFIDFAttackModelBuilderV1 behavior by extracting requests from a
     * text file; calculating the TF-IDF based on the default model (read from
     * the data directory); and outputting the resulting classifications -- one
     * per request.
     * 
     * Arguments: 
     *   args[0] = HDFS directory of input file(s) containing requests
     *              NOTE: Input files are expected to contain an attack type 
     *              followed by all HTTP requests associated with it 
     *   args[1] = HDFS file (directory + filename) for model file 
     *   args[2] = Boolean indicating whether the first (tab-delimited) field 
     *             of each line of the input file(s) is a request identifier 
     *             (which must be ignored during the model building process).
     */
    public int run(String[] args) throws Exception {
        int res = 0;

        if (args.length == 3) {
            buildInMemoryModel(args[0], args[1], args[2]);
        } else {
            System.out.println(g_lt_myName
                    + " <inDir> <modelFile> <stripRequestIds>");
            System.out.println("                  "
                    + "where the first two arguments exist in the HDFS.");
            System.out.println("                  "
                    + "Also <stripRequestIds> should be 'true' if each");
            System.out.println("                  "
                    + "line of the input file(s) start with the associated");
            System.out.println("                  "
                    + "identifier followed by a tab.");
            System.out.println();
            ToolRunner.printGenericCommandUsage(System.out);
            res = -1;
        }

        return res;
    } // run(String[]): int

    /* ***************************************************************** */
    /* main() */
    /* ***************************************************************** */
    // @SuppressWarnings("unchecked")
    public static void main(String[] args) throws Exception {
        int res = ToolRunner.run(new Configuration(), 
                                 new TFIDFAttackModelBuilderV1(), args);
        System.exit(res);
    } // main(String[])

} // TFIDFAttackModelBuilderV1
