/**
 * Name   :  TFIDFAttackModelV1.java
 * Purpose:  Manage a Term-Frequency, Inverse Document Frequency (TF-IDF) 
 *           model for HTTP attacks.  The official/general TF-IDF formulas are:
 *          
 *             tf   = n  / Sum n    = count(t , d )/ Sum      count(v, d )
 *               ij    ij   k   kj           i   j     v in d           j
 *            
 *             idf  = log (|D|)/|{d : t in d }|
 *                i                j   i    j
 *          
 *           so tfidf   = tf   * idf
 *                   ij     ij      i
 *                  
 *           where d = request, D = training set (i.e., corpus), t and v = term,
 *           tf = term frequency, idf = inverse document frequency, and 
 *           i,j,k are counters.
 *          
 *           In this implementation, |D| = the number of attack types since
 *           the source (see Acknowledgements) treated all requests for each
 *           attack type as a separate document.
 * Author :  T. Dahlgren
 * Created:  December 19, 2008
 *
 * Acknowledgments:  
 *   This class mimics Brian Gallagher's non-Hadoop model.
 *   
 * Assumptions:
 *   1) Keys to HashMaps are simple Java types (e.g., String) versus objects
 *      so they are unique within the maps.
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package gov.llnl.hadoop.classifier.httpRequest;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.hadoop.io.WritableComparable;

public class TFIDFAttackModelV1 implements WritableComparable {
    /* ====================== ATTRIBUTES ====================== */
    /*
     * d_distribution  = Default probability distributions as <attackType,
     *                   probability>. 
     * d_termTFIDFs    = TF-IDF vectors as <term, <attackType, TFIDF>>.
     * d_idfs          = IDF vectors as <term, IDF>. 
     * d_attackNorms   = TF-IDF vector norms as <attackType, ||TFIDF|| 
     * d_initialized   = TRUE if the attack model has been initialized. 
     * d_format        = Format used when outputting decimal numbers.
     * 
     * For simplicity, mirroring the attribute ordering in Brian's code to
     * reflect the ordering in his model. (The expectation/hope is to re-use 
     * his model for initial testing.)
     */
    private Map<String, Double> d_distribution = null;
    private Map<String, Map<String, Double>> d_termTFIDFs = null;
    private Map<String, Double> d_idfs = null;
    private Map<String, Double> d_attackNorms = null;
    private boolean d_initialized = false;

    private static final DecimalFormat d_format = new DecimalFormat("0.000000");

    /* ====================== METHODS ====================== */
    public TFIDFAttackModelV1() {
    } // TFIDFAttackModelV1()

    public TFIDFAttackModelV1(DataInput in) {
        try {
            restoreModel(in);
        } catch (Exception exc) {
            System.err.println("Cannot instantiate model from file.");
            exc.printStackTrace();
        }
    } // TFIDFAttackModelV1(String)

    @SuppressWarnings("unchecked")
    // For unchecked casts
    private void restoreModel(DataInput in) throws IOException {
        try {
            ObjectInputStream ois = new ObjectInputStream((InputStream) in);

            d_distribution = (Map<String, Double>) ois.readObject();
            d_termTFIDFs = (Map<String, Map<String, Double>>) ois.readObject();
            d_idfs = (Map<String, Double>) ois.readObject();
            d_attackNorms = (Map<String, Double>) ois.readObject();
            d_initialized = true;
            ois.close();
        } catch (ClassNotFoundException cnfe) {
            System.err
                    .println("Cannot restore model objects from input stream "
                            + "due to ClassNotFoundException");
            cnfe.printStackTrace();
        }
    } // restoreModel(InputStream)

    /*
     * Tokenize the request.
     */
    private String[] tokenizeRequest(String request) {
        String[] termList = null;

        /*
         * Tokenize the request after removing plus signs and hex code
         * characters. (In other words, remove the same special characters Brian
         * Gallagher removed in his implementation.) Also ensure all characters
         * are in lower case to facilitate comparisons.
         * 
         * TBD: Shouldn't this really convert some of the hex characters since,
         * unless I'm misremembering, they may be part of an attack?
         */
        if (request != null) {
            request = request.replaceAll("\\+", " ");
            request = request.replaceAll("%..", " ");
            termList = request.toLowerCase().split("\\s+");
        }

        return termList;
    } // tokenizeRequest(String): String[]

    /*
     * Build the attack model from a training set WITHOUT using map/reduce.
     * 
     * Each requestSet entry is assumed to take the form <attackType, requests>
     * to match Brian's pre-processing of the XML training set. That is, the key
     * is assumed to be the attack type label and the value, or requests, is
     * assumed to be one or more HTTP requests depending on the number present
     * in the training set.
     */
    public void buildModel(Map<String, String> requestSet) {
        Map<String, Map<String, Long>> attackTermCounts = new HashMap<String, Map<String, Long>>();
        String attackType;

        /*
         * Calculate the attack distribution.
         */
        d_distribution = new HashMap<String, Double>();
        double attackProb = 1.0 / requestSet.size();
        Map<String, Long> totalAttackTerms = new HashMap<String, Long>();
        for (Map.Entry<String, String> req : requestSet.entrySet()) {
            attackType = req.getKey();
            String attackRequests = req.getValue();
            String[] terms = tokenizeRequest(attackRequests);

            /*
             * Initialize/accumulate total terms for the attack type (i.e.,
             * totalAttackTerms is <attackType, totalTerms>).
             */
            Long count = totalAttackTerms.get(attackType);
            long cVal = 0;
            if (count != null) {
                cVal = count.longValue();
            }
            totalAttackTerms.put(attackType, new Long(cVal + terms.length));

            /*
             * Initialize/accumulate term totals for the attack type (i.e.,
             * termCounts is <term, termCount> BY attackType).
             */
            Map<String, Long> termCounts = attackTermCounts.get(attackType);
            if (termCounts == null) {
                termCounts = new HashMap<String, Long>();
            }
            for (String term : terms) {
                Long tCount = termCounts.get(term);
                long tcVal = 0;
                if (tCount != null) {
                    tcVal = tCount.longValue();
                }
                termCounts.put(term, new Long(tcVal + 1));
            } // endfor each term in the attack's request(s)
            attackTermCounts.put(attackType, termCounts);

            /*
             * Initialize/accumulate attack probabilities.
             */
            Double prob = d_distribution.get(attackType);
            double pVal = 0.0;
            if (prob != null) {
                pVal = prob.doubleValue();
            }
            d_distribution.put(attackType, new Double(pVal + attackProb));
        } // endfor each attack requestSet entry

        /*
         * Calculate TF-IDF vectors for each term.
         */
        d_termTFIDFs = new HashMap<String, Map<String, Double>>();
        for (Map.Entry<String, Map<String, Long>> counts : attackTermCounts
                .entrySet()) {
            attackType = counts.getKey();

            /*
             * Initialize/accumulate attack TFs.
             */
            for (Map.Entry<String, Long> tCount : counts.getValue().entrySet()) {
                String term = tCount.getKey();

                /*
                 * Normalize the count by the total number of terms appearing in
                 * the attack request(s).
                 */
                Map<String, Double> attackTF = d_termTFIDFs.get(term);
                if (attackTF == null) {
                    attackTF = new HashMap<String, Double>();
                }

                double prob = (double) (tCount.getValue().longValue())
                    / (double) (totalAttackTerms.get(attackType).longValue());
                attackTF.put(attackType, new Double(prob));
                d_termTFIDFs.put(term, attackTF);
            } // endfor each attack associated with the term
        } // endfor each attack type

        /*
         * Calculate term IDF vector and attack norms.
         */
        d_idfs = new HashMap<String, Double>();
        d_attackNorms = new HashMap<String, Double>();
        int numAttackTypes = d_distribution.size();
        Iterator<String> termIter = d_termTFIDFs.keySet().iterator();
        while (termIter.hasNext()) {
            String term = termIter.next();
            Map<String, Double> attackTF = d_termTFIDFs.get(term);
            if (attackTF != null) {
                /*
                 * Count the total number of attack type request sets in which
                 * the term appears.
                 */
                long termAttackTotal = 0;
                for (Double tfs : attackTF.values()) {
                    if (tfs.doubleValue() > 0.0) {
                        termAttackTotal++;
                    }
                } // endfor each TFIDF

                /*
                 * Calculate TF-IDF for an attack type. Also accumulate attack
                 * norm sum.
                 */
                double idfWeight = 0.0;
                if (termAttackTotal > 0) {
                    idfWeight = Math.log((double)numAttackTypes 
                                         / (double)termAttackTotal);
                }
                if (idfWeight != 0.0) {
                    d_idfs.put(term, idfWeight);
                    for (String aType : attackTF.keySet()) {
                        Double tf = attackTF.get(aType);
                        Double tfidf = tf * idfWeight;
                        attackTF.put(aType, tfidf);

                        Double normSum = d_attackNorms.get(aType);
                        double nsVal = 0.0;
                        if (normSum != null) {
                            nsVal = normSum.doubleValue();
                        }
                        normSum = new Double(nsVal + tfidf * tfidf);
                        d_attackNorms.put(aType, normSum);
                    } // endfor each attack type with a TFIDF entry
                } else {
                    /*
                     * Remove TFIDF entries if inapplicable.
                     */
                    termIter.remove();
                } // endif potential attack type
            } // endif attack TFIDF entry for the term
        } // endwhile terms with TFIDF vectors

        /*
         * Complete calculation of attack norm vectors.
         */
        for (String aType : d_attackNorms.keySet()) {
            d_attackNorms.put(aType, Math.sqrt(d_attackNorms.get(aType)
                    .doubleValue()));
        } // endfor each attack type

        d_initialized = true;
    } // buildModel(HashMap)

    /*
     * Calculate cosine similarity vector of the request passed as a list of
     * terms.
     */
    private Map<String, Double> calcCosineSimilarity(String[] requestTerms) {
        Map<String, Double> csVec = null;

        if ((requestTerms.length > 0) && d_initialized) {
            csVec = new HashMap<String, Double>();

            /*
             * Calculate TF for each unique terms in the request.
             */
            Map<String, Double> tfs = new HashMap<String, Double>();
            double frac = 1.0 / requestTerms.length;
            for (String term : requestTerms) {
                Double freq = tfs.get(term);
                double fVal = 0.0;
                if (freq != null) {
                    fVal = freq.doubleValue();
                }
                tfs.put(term, new Double(fVal + frac));
            } // endfor each term in the request

            /*
             * Calculate the dot products of each unique term.
             */
            double reqNormSum = 0.0;
            for (String term : tfs.keySet()) {
                Map<String, Double> attackTFIDFs = d_termTFIDFs.get(term);
                if (attackTFIDFs != null) {
                    /*
                     * Accumulate TFIDF for each attack type associated with the
                     * term.
                     */
                    double termTFIDF = tfs.get(term).doubleValue()
                            * d_idfs.get(term).doubleValue();
                    for (Map.Entry<String, Double> e : attackTFIDFs.entrySet()) {
                        String attackType = e.getKey();
                        Double tfidf = csVec.get(attackType);
                        double tfidfVal = 0.0;
                        if (tfidf != null) {
                            tfidfVal = tfidf.doubleValue();
                        }
                        double tfidfComp = e.getValue().doubleValue()
                                * termTFIDF;
                        csVec.put(attackType, new Double(tfidfVal + tfidfComp));
                    } // endfor each attack type in the term's TFIDF

                    /*
                     * Accumulate values for calculating the TFIDF norm.
                     */
                    reqNormSum += termTFIDF * termTFIDF;
                } // endif have a TFIDF vector for the term
            } // endfor each unique term

            /*
             * Complete the request's TFIDF norm calculation.
             */
            double reqNorm = Math.sqrt(reqNormSum);

            /*
             * Calculate the resulting distributions for each attack type.
             */
            for (String attackType : csVec.keySet()) {
                double freq = csVec.get(attackType).doubleValue()
                        / (d_attackNorms.get(attackType).doubleValue() * reqNorm);
                csVec.put(attackType, new Double(freq));
            } // endfor each attack type
        } // endif have terms to process in the request

        return csVec;
    } // calcCosineSimilarity(String[]): Map

    /*
     * Calculate the attack type distribution for the given request.
     */
    public Map<String, Double> calcDistribution(String request) {
        Map<String, Double> dist = null;

        if ((request != null) && d_initialized) {
            String[] terms = tokenizeRequest(request);
            dist = calcCosineSimilarity(terms);

            /*
             * Normalize attack type probabilities (so sum to 1.0).
             */
            double totalProb = 0.0;
            for (String attackType : dist.keySet()) {
                totalProb += dist.get(attackType).doubleValue();
            }
            for (String attackType : dist.keySet()) {
                dist.put(attackType, new Double(dist.get(attackType)
                        .doubleValue()
                        / totalProb));
            }

            /*
             * Ensure request's distribution includes entries for all known
             * attack types from the model's distribution.
             */
            if (!dist.isEmpty()) {
                for (String attackType : d_distribution.keySet()) {
                    if (!dist.containsKey(attackType)) {
                        dist.put(attackType, new Double(0.0));
                    }
                }
            } else {
                dist = d_distribution;
            }
        } else if (!d_initialized) {
            System.err.println("Cannot calculate distribution before "
                    + "model has been initialized.");
        } else {
            System.err.println("Cannot calculate distribution for an "
                    + "empty request.");
        } // endif have a request to process

        return dist;
    } // calcDistribution(request): Map

    public void readFields(DataInput in) throws IOException {
        restoreModel(in);
    } // readFields(DataInput)

    public void write(DataOutput out) throws IOException {
        ObjectOutputStream oos = new ObjectOutputStream((OutputStream) out);
        try {
            oos.writeObject(d_distribution);
            oos.writeObject(d_termTFIDFs);
            oos.writeObject(d_idfs);
            oos.writeObject(d_attackNorms);
        } finally {
            oos.close();
        }
    } // write(DataOutput)

    /*
     * Append the description, desc, and contents of the map, m, to the buffer,
     * buff.
     */
    private void appendMap(String desc, String tab, Map<String, Double> m,
            StringBuffer buff) {
        if (desc != null) {
            String emptyStr = (m.size() > 0) ? "" : "  Is empty";
            buff.append("\n" + desc + ":" + emptyStr + "\n");
        }
        for (Map.Entry<String, Double> e : m.entrySet()) {
            buff.append(tab + e.getKey() + ": " + d_format.format(e.getValue())
                    + "\n");
        }
    } // appendMap(String, String, Map, StringBuffer)

    /*
     * Stringify model contents.
     */
    public String toString() {
        StringBuffer str = new StringBuffer();
        if (d_initialized) {
            appendMap("Distribution", "  ", d_distribution, str);
            String emptyStr = (d_termTFIDFs.size() > 0) ? "" : "  Is empty";
            str.append("\nTerm TFIDFs: " + emptyStr + "\n");
            for (Map.Entry<String, Map<String, Double>> e : d_termTFIDFs
                    .entrySet()) {
                appendMap("  " + e.getKey(), "    ", e.getValue(), str);
            }
            appendMap("\nIDFs", "  ", d_idfs, str);
            appendMap("\nAttack Norms", "  ", d_attackNorms, str);
        }
        return str.toString();
    } // toString(): String

    /*
     * Comparison function to allow object sorting (as in secondary sort!) Don't
     * really have anything worth comparing here, so simply returning 0 to
     * indicate relative order in "sorted" list does not matter.
     */
    public int compareTo(Object obj) {
        return 0;
    } // compareTo(Object): int
} // TFIDFAttackModelV1
